/********** U T I L I T Y ************************************
 * Utility functions for use with the Jacobi and SOR solvers *
 * Kadin Tseng, Boston University, November 1999             *
 *************************************************************/

#include "solvers.h"

REAL **allocate_2D(INT m, INT n) {
   INT i;
   REAL **a;

   a = (REAL **) malloc((unsigned) m*sizeof(REAL*));

/* Each pointer array element points to beginning of a row with n entries*/
   for (i = 0; i < m; i++) {
     a[i] = (REAL *) malloc((unsigned) n*sizeof(REAL));
   }

   return a;
}

INT write_file( INT m, INT n, REAL **u, INT k, INT p ) {
/*****************************************************
 * Writes 2D array ut columnwise (i.e. C convention) *
 * m   - size of rows                                *
 * n   - size of columns                             *
 * u   - scratch array                               *
 * k   - 0 <= k < p; = 0 for single thread code      *
 * p   - p >= 0; =1 for single thread code           *
 *****************************************************/
  INT ij, i, j, per_line;
  CHAR filename[50], file[53];
  FILE *fd;
/*
   prints u, 6 per line; used for matlab plots;
   PLOT_FILE contains the array size and number of procs;
   PLOT_FILE.(k+1) contains u pertaining to proc k;
   for serial job, PLOT_FILE.1 contains full u array.
*/

   (void) sprintf(filename, "%s", PLOT_FILE);

   if ( k == 0 ) {
      fd = fopen(filename, "w");
      fprintf(fd, "%5d %5d %5d\n", m, n, p);
      fclose(fd);
   }

   per_line = 6;                        /* to print 6 per line */
   (void) sprintf(file, "%s.%d", filename, k); /* create output file */
   fd = fopen(file, "w");
   ij = 0;
   for (j = 0; j < n; j++) {
     for (i = 0; i < m; i++) {
       fprintf(fd, "%11.4f ", u[i][j]);
       if ((ij+1)%per_line == 0) fprintf(fd, "\n");
       ij++;
     }
   }
   fprintf(fd, "\n");
   fclose(fd);
   return (0);
}


INT write_gnuplot( INT m, INT n, REAL **u) {
  INT i, j;
  FILE* fd;

  fd = fopen(PLOT_FILE, "w");

   for (j = 0; j < n; j++) {
     for (i = 0; i < m; i++) {
       fprintf(fd, "%f %f %11.4f\n", 1.0/(m-1)*i, 1.0/(n-1)*j, u[i][j]);
     }
     fprintf(fd, "\n");     
   }
   fclose(fd);
   return (0);
}

void init_array(INT m, INT n, REAL **a) {
/********* Initialize Array **********************
 * Initialize array with nx rows and ny columns  *
 *************************************************/
  INT i, j;

  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      a[i][j] = 0.0;          /* initialize all entries to zero */
    }
  }
}

void bc(INT m, INT n, REAL **u, INT k, INT p) {
/*********** Boundary Conditions ****************************************
 *  PDE: Laplacian u = 0;      0<=x<=1;  0<=y<=1                        *
 *  B.C.: u(x,0)=sin(pi*x); u(x,1)=sin(pi*x)*exp(-pi); u(0,y)=u(1,y)=0  *
 *  SOLUTION: u(x,y)=sin(pi*x)*exp(-pi*y)                               *
 ************************************************************************/
  INT i;

  init_array( m, n, u);                         /* initialize u to 0 */

  if (p > 1) {
    if (k == 0) {
      for (i = 0; i < m; i++) {
        u[i][0] = sin(PI*i/(m-1));              /* at y = 0; all x */
      }
    }
    if (k == p-1) {
      for (i = 0; i < m; i++) {
        u[i][n-1] = sin(PI*i/(m-1))*exp(-PI);   /* at y = 1; all x */
      }
    }
  } else if (p == 1) {
      for (i = 0; i < m; i++) {
        u[i][  0] = sin(PI*i/(m-1));    /* at y = 0; all x */
        u[i][n-1] = u[i][0]*exp(-PI);   /* at y = 1; all x */
      }
  } else {
    printf("p is invalid\n");
  }
}

void prtarray( INT m, INT n, REAL **a, FILE *fd) {
/*********** Print Array ***********************
 * Prints array "a" with m rows and n columns  *
 * tda is the Trailing Dimension of Array a    *
 ***********************************************/
  INT i, j;
  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      fprintf(fd, "%8.2f", a[i][j]);
    }
    fprintf(fd, "\n");
  }
}
  
INT update_jacobi( INT m, INT n, REAL **u, REAL **unew, REAL *del) {
/**********************************************************************
 * Updates u according to Jacobi method                               *
 * m        - (INPUT)  size of interior rows                          *
 * n        - (INPUT)  size of interior columns                       *
 * u        - (INPUT)  solution array                                 *
 * unew     - (INPUT)  next solution array                            *
 * del      - (OUTPUT) error norm between 2 solution steps            *
 **********************************************************************/
  INT i, j;
  *del = 0.0;
  for (i = 1; i < m-1; i++) {
    for (j = 1; j < n-1; j++) {
      unew[i][j] = ( u[i  ][j+1] + u[i+1][j  ] +
                     u[i-1][j  ] + u[i  ][j-1] )*0.25;
      *del += fabs(unew[i][j] - u[i][j]);    /* find local max error */
    }
  }
  for (i = 1; i < m-1; i++) {
    for (j = 1; j < n-1; j++) {
      u[i][j] = unew[i][j];
    }
  }
      
  return (0);
}

INT update_sor( INT m, INT n, REAL **u, REAL omega, REAL *del, CHAR redblack) {
/**********************************************************************
 * Updates u according to successive over relaxation method           *
 * m        - (INPUT)  size of interior rows                          *
 * n        - (INPUT)  size of interior columns                       *
 * u        - (INPUT)  array                                          *
 * omega    - (INPUT)  adjustable constant used to speed up           *
 *                     convergence of SOR                             *
 * del      - (OUTPUT) error norm between 2 solution steps            *
 * redblack - (INPUT)  either 'r' for red and 'b' for black           *
 **********************************************************************/
   INT i, ib, ie, j, jb, je;
   REAL up;

      *del = 0.0;
      if (redblack == 'r') {
/* process RED odd points ... */
        jb = 1; je = n-2; ib = 1; ie = m-2;
        for ( j = jb; j <= je; j+=2 ) {
          for ( i = ib; i <=ie; i+=2 ) {
            up = ( u[i  ][j+1] + u[i+1][j  ] +
                   u[i-1][j  ] + u[i  ][j-1] )*0.25;
            u[i][j] = (1.0 - omega)*u[i][j] + omega*up;
            *del += fabs(up-u[i][j]);
          }
        }
/* process RED even points ... */
        jb = 2; je = n-2; ib = 2; ie = m-2;
        for ( j = jb; j <= je; j+=2 ) {
          for ( i = ib; i <= ie; i+=2 ) {
            up = ( u[i  ][j+1] + u[i+1][j  ] +
                   u[i-1][j  ] + u[i  ][j-1] )*0.25;
            u[i][j] = (1.0 - omega)*u[i][j] + omega*up;
            *del += fabs(up-u[i][j]);
          }
        }
        return (0);
      } else { 
          if (redblack == 'b') {
/* process BLACK odd points ... */
            jb = 2; je = n-2; ib = 1; ie = m-2;
            for ( j = jb; j <= je; j+=2 ) {
              for ( i = ib; i <= ie; i+=2 ) {
                up = ( u[i  ][j+1] + u[i+1][j  ] +
                       u[i-1][j  ] + u[i  ][j-1] )*0.25;
                u[i][j] = (1.0 - omega)*u[i][j] + omega*up;
                *del += fabs(up-u[i][j]);
              }
            }
/* process BLACK even points ... */
            jb = 1; je = n-2; ib = 2; ie = m-2;
            for ( j = jb; j <= je; j+=2 ) {
              for ( i = ib; i <= ie; i+=2 ) {
                up = ( u[i  ][j+1] + u[i+1][j  ] +
                       u[i-1][j  ] + u[i  ][j-1] )*0.25;
                u[i][j] = (1.0 - omega)*u[i][j] + omega*up;
                *del += fabs(up-u[i][j]);
              }
            }
            return (0);
          } else {
            return (1);
          }
      }
}

INT replicate( INT m, INT n, REAL **a, REAL **b ) {
/********************************************************
 * Replicates array a into array b                      *
 * m    - (INPUT)  size of interior points in 1st index *
 * n    - (INPUT)  size of interior points in 2st index *
 * a    - (INPUT)  solution at time N                   *
 * b    - (OUTPUT) solution at time N + 1               *
 ********************************************************/
  INT i, j;

  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      b[i][j] = a[i][j];
    }
  }
  return (0);
}

INT transpose( INT m, INT n, REAL **a, REAL **at ) {
/**********************************************************
 * Transpose a(0:m+1,0:n+1) into at(0:n+1,0:m+1)          *
 * m    - (INPUT)  size of interior points in 1st index   *
 * n    - (INPUT)  size of interior points in 2st index   *
 * a    - (INPUT)  a  =  a(0:m+1,0:n+1)                   *
 * at   - (OUTPUT) at = at(0:n+1,0:m+1)                   *
 **********************************************************/
  INT i, j;

  for (i = 0; i < m; i++) {
    for (j = 0; j < n; j++) {
      at[j][i] = a[i][j];
    }
  }
  return (0);
}

void neighbors(INT k, INT p, INT UNDEFINED, INT *below, INT *above) {
/****************************************************************
 * determines two adjacent threads                              *
 * k         - (INPUT)  current thread                          *
 * p         - (INPUT)  number of processes (threads)           *
 * UNDEFINED - (INPUT)  code to assign to out-of-bound neighbor *
 * below     - (OUTPUT) neighbor thread below k (usually k-1)   *
 * above     - (OUTPUT) neighbor thread above k (usually k+1)   *
 ****************************************************************/
  if(k == 0) {
    *below = UNDEFINED;     /* tells MPI not to perform send/recv */
    *above = k+1;
  } else if(k == p-1) {
    *below = k-1;
    *above = UNDEFINED;     /* tells MPI not to perform send/recv */
  } else {
    *below = k-1;
    *above = k+1;
  }
}

